/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SBS_CONE_DEBUG	FALSE

#include "surface.pmh"

namespace fe
{

SurfaceI::Containment SurfaceBase::containment(
	const SpatialVector& a_origin) const
{
	sp<SurfaceI::ImpactI> spImpact=nearestPoint(a_origin);
	if(!spImpact.isValid())
	{
		return SurfaceI::e_unknown;
	}

	const SpatialVector intersection=spImpact->intersection();
	const SpatialVector outward=spImpact->normal();
	const SpatialVector toward=intersection-a_origin;

	//* no need to normalize vs zero
	return (dot(outward,toward)<0.0)? SurfaceI::e_outside: SurfaceI::e_inside;
}

sp<SurfaceI::ImpactI> SurfaceBase::coneImpact(const SpatialVector& a_origin,
		const SpatialVector& a_direction,Real a_maxDistance,
		Real a_coneAngle, sp<DrawI> a_spDrawI)
{
	sp<SurfaceI::ImpactI> spResult=rayImpact(a_origin,a_direction,
			a_maxDistance);
	const SpatialVector different(a_direction[1],a_direction[2],
			-a_direction[0]);

	SpatialVector alignment;
	cross(alignment,a_direction,different);

#if FE_SBS_CONE_DEBUG
	feLog("\ncast %s\n",c_print(a_origin));
	feLog("dir %s\n",c_print(a_direction));
	if(spResult.isValid())
	{
		feLog("dist %.6G\n",spResult->distance());
	}
#endif

	const U32 count=a_coneAngle>0.0? 6: 0;	//* tweak
	for(U32 m=0;m<count;m++)
	{
		SpatialQuaternion spin(2.0*M_PI*m/Real(count),a_direction);
		SpatialVector axis;
		rotateVector(spin,alignment,axis);

		SpatialQuaternion splay(a_coneAngle,axis);
		SpatialVector redirection;
		rotateVector(splay,a_direction,redirection);

#if FE_SBS_CONE_DEBUG
		feLog("%d axis %s\n",m,c_print(axis));
		feLog("redir %s\n",c_print(redirection));
#endif


		sp<SurfaceI::ImpactI> spCandidate=rayImpact(a_origin,redirection,
				a_maxDistance);
#if FE_SBS_CONE_DEBUG
		if(spCandidate.isValid())
		{
			feLog("dist %.6G\n",spCandidate->distance());
		}
#endif
		if(a_spDrawI.isValid())
		{
			const Color blue(0.0,0.0,1.0,1.0);
			const Color lightblue(0.5,0.5,1.0,1.0);
			SpatialVector line[2];
			line[0]=a_origin;
			line[1]=a_origin+redirection;
			a_spDrawI->drawLines(line,NULL,2,
					DrawI::e_discrete,false,
					(spCandidate.isValid() && spCandidate->distance()>0.0)?
					&blue: &lightblue);
		}

		if(!spCandidate.isValid() || spCandidate->distance()<0.0)
		{
			continue;
		}
		if(!spResult.isValid() || spResult->distance()<0.0 ||
				spCandidate->distance()<spResult->distance())
		{
			spResult=spCandidate;
#if FE_SBS_CONE_DEBUG
			feLog("REPLACE\n");
#endif
		}
	}
	return spResult;
}

sp<SurfaceI::ImpactI> SurfaceBase::coneImpact(
	const SpatialTransform& a_transform,
	const SpatialVector& a_origin,const SpatialVector& a_direction,
	Real a_maxDistance,Real a_coneAngle,sp<DrawI> a_spDrawI)
{
	const SpatialVector invOrigin=inverseTransformVector(a_transform,a_origin);
	const SpatialVector invDirection=
			inverseRotateVector(a_transform,a_direction);
	sp<Impact> spImpact=coneImpact(invOrigin,invDirection,a_maxDistance,
			a_coneAngle,a_spDrawI);
	if(spImpact.isValid())
	{
		spImpact->setTransform(a_transform);
	}
	return spImpact;
}

} // namespace
