/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

#define	FE_ST_TREE_DRAW_DEBUG	FALSE
#define	FE_ST_TREE_DRAW_NORMALS	FALSE

namespace fe
{

SurfaceTriangles::SurfaceTriangles(void)
{
	m_pImpactPool=&m_triangleImpactPool;
}

SurfaceTriangles::~SurfaceTriangles(void)
{
}

sp<SurfaceI::ImpactI> SurfaceTriangles::sampleImpact(I32 a_triangleIndex,
	SpatialBary a_barycenter,SpatialVector a_tangent) const
{
	const SpatialTransform xform=
			sample(a_triangleIndex,a_barycenter,a_tangent);

	//* TODO clean up duplicate code with sample()
	const U32 faceCount=m_vertices/3;
	I32 face;
	for(face=0;face<faceCount;face++)
	{
		if(m_pTriangleIndexArray[face*3]==a_triangleIndex)
		{
			break;
		}
	}

	const I32 face3=face*3;
	if(face3+3>m_vertices)
	{
		return sp<Impact>(NULL);
	}

	const SpatialVector& rVertex0=m_pVertexArray[face3];
	const SpatialVector& rVertex1=m_pVertexArray[face3+1];
	const SpatialVector& rVertex2=m_pVertexArray[face3+2];

	SpatialVector du;
	SpatialVector dv;
	set(du);
	set(dv);
	if(m_pUVArray)
	{
		triangleDuDv(du,dv,rVertex0,rVertex1,rVertex2,
				m_pUVArray[face3],m_pUVArray[face3+1],m_pUVArray[face3+2]);
	}

	sp<Impact> spImpact=m_pImpactPool->get();

	spImpact->setSurface(dynamic_cast<const SurfaceTriangles*>(this));

	spImpact->setTriangleIndex(a_triangleIndex);
	spImpact->setBarycenter(a_barycenter);
	spImpact->setDu(du);
	spImpact->setDv(dv);

	spImpact->setPointIndex0(m_pPointIndexArray[face3]);
	spImpact->setPointIndex1(m_pPointIndexArray[face3+1]);
	spImpact->setPointIndex2(m_pPointIndexArray[face3+2]);

	spImpact->setVertex0(rVertex0);
	spImpact->setVertex1(rVertex1);
	spImpact->setVertex2(rVertex2);

	spImpact->setNormal0(m_pNormalArray[face3]);
	spImpact->setNormal1(m_pNormalArray[face3+1]);
	spImpact->setNormal2(m_pNormalArray[face3+2]);

	const SpatialVector intersection=xform.translation();
	const Real distance=magnitude(intersection);
	spImpact->setOrigin(SpatialVector(0.0,0.0,0.0));
	spImpact->setDistance(distance);
	spImpact->setDirection(intersection*(distance>0.0? 1.0/distance: 1.0));

	return spImpact;
}

SpatialTransform SurfaceTriangles::sample(I32 a_triangleIndex,
	SpatialBary a_barycenter,SpatialVector a_tangent) const
{
	//* TODO map
	const U32 faceCount=m_vertices/3;
	I32 face;
	for(face=0;face<faceCount;face++)
	{
		if(m_pTriangleIndexArray[face*3]==a_triangleIndex)
		{
			break;
		}
	}

	const I32 face3=face*3;
	if(face3+3>m_vertices)
	{
#if FALSE
		//* WARN multi-threaded logging can segfault
		feLog("SurfaceTriangles::sample"
				" excess face %d (%d>%d) face %d/%d\n",
				a_triangleIndex,face3+3,m_vertices,face,faceCount);
#endif
		return SpatialTransform::identity();
	}
	const SpatialVector& rVertex0=m_pVertexArray[face3];
	const SpatialVector& rVertex1=m_pVertexArray[face3+1];
	const SpatialVector& rVertex2=m_pVertexArray[face3+2];
	const SpatialVector& rNormal0=m_pNormalArray[face3];
	const SpatialVector& rNormal1=m_pNormalArray[face3+1];
	const SpatialVector& rNormal2=m_pNormalArray[face3+2];

	SpatialVector point;
	SpatialVector norm;
	if(m_sampleMethod==e_pointNormal)
	{
		//* TODO can we cache the configure?

		TrianglePN<Real> trianglePN;
		trianglePN.configure(rVertex0,rVertex1,rVertex2,
				rNormal0,rNormal1,rNormal2);
		trianglePN.solve(a_barycenter,point,norm);
	}
	else
	{
		point=location(a_barycenter,rVertex0,rVertex1,rVertex2);
		norm=location(a_barycenter,rNormal0,rNormal1,rNormal2);
	}

	Real mag=magnitude(norm);
	SpatialVector tangent=a_tangent;
	SpatialVector du;
	SpatialVector dv;
	set(du);
	set(dv);

	SpatialTransform result;
	if(mag>0.0)
	{
		if(m_pUVArray)
		{
#if TRUE
			triangleDuDv(du,dv,rVertex0,rVertex1,rVertex2,
					m_pUVArray[face3],m_pUVArray[face3+1],m_pUVArray[face3+2]);

			tangent=unitSafe(du+dv);
#else
			const SpatialVector uv=location(a_barycenter,
					SpatialVector(m_pUVArray[face3]),
					SpatialVector(m_pUVArray[face3+1]),
					SpatialVector(m_pUVArray[face3+2]));

			SpatialVector du0;
			SpatialVector dv0;
			triangleDuDv(du0,dv0,point,rVertex1,rVertex2,
					uv,m_pUVArray[face3+1],m_pUVArray[face3+2]);

			SpatialVector du1;
			SpatialVector dv1;
			triangleDuDv(du1,dv1,rVertex0,point,rVertex2,
					m_pUVArray[face3],uv,m_pUVArray[face3+2]);

			SpatialVector du2;
			SpatialVector dv2;
			triangleDuDv(du2,dv2,rVertex0,rVertex1,point,
					m_pUVArray[face3],m_pUVArray[face3+1],uv);

			du=unitSafe(du0+du1+du2);
			dv=unitSafe(dv0+dv1+dv2);

			tangent=unitSafe(du+dv);
#endif
		}
		if(magnitudeSquared(tangent)<1e-6)
		{
			tangent=unitSafe(unitSafe(rVertex1-rVertex0)+
					unitSafe(rVertex2-rVertex0));
		}
		makeFrameNormalY(result,point,tangent,norm/mag);

//		feLog("a_tangent %s tangent %s\n",c_print(a_tangent),c_print(tangent));
	}
	else
	{
		setIdentity(result);
		setTranslation(result,point);
	}

#if FALSE
	feLog("tri %d face %d/%d bary %s\n",
			a_triangleIndex,face,faceCount,c_print(a_barycenter));
	feLog("rVertex0 %s\n",c_print(rVertex0));
	feLog("rVertex1 %s\n",c_print(rVertex1));
	feLog("rVertex2 %s\n",c_print(rVertex2));
	feLog("rNormal0 %s\n",c_print(rNormal0));
	feLog("rNormal1 %s\n",c_print(rNormal1));
	feLog("rNormal2 %s\n",c_print(rNormal2));
	feLog("du %s\n",c_print(du));
	feLog("dv %s\n",c_print(dv));
	if(m_pUVArray)
	{
		feLog("uv0 %s\n",c_print(m_pUVArray[face3]));
		feLog("uv1 %s\n",c_print(m_pUVArray[face3+1]));
		feLog("uv2 %s\n",c_print(m_pUVArray[face3+2]));
	}
	feLog("tangent %s\n",c_print(tangent));
	feLog("point %s\n",c_print(point));
	feLog("norm %s\n",c_print(norm));
	feLog("result %s\n",c_print(result));
#endif

	return result;
}

void SurfaceTriangles::drawInternal(BWORD a_transformed,
	const SpatialTransform* a_pTransform,sp<DrawI> a_spDrawI,
	const fe::Color* a_pColor) const
{
//	return;

#if FE_ST_TREE_DRAW_DEBUG
	sp<DrawMode> spWireframe(new DrawMode);
	spWireframe->copy(a_spDrawI->drawMode());
	spWireframe->setDrawStyle(DrawMode::e_outline);
	a_spDrawI->pushDrawMode(spWireframe);
#endif

#if TRUE
	if(m_vertices)
	{
		const Color white(1.0f,1.0f,1.0f,1.0f);

		const Vector2f* pTexture=NULL;
		const BWORD multicolor=FALSE;
		if(a_transformed)
		{
			a_spDrawI->drawTransformedTriangles(*a_pTransform,NULL,
					m_pVertexArray,m_pNormalArray,pTexture,m_vertices,
					DrawI::e_discrete,multicolor,a_pColor? a_pColor: &white);
		}
		else
		{
			a_spDrawI->drawTriangles(m_pVertexArray,m_pNormalArray,pTexture,
					m_vertices,DrawI::e_discrete,multicolor,
					a_pColor? a_pColor: &white);
		}
	}
	else
	{
		feLog("SurfaceTriangles::draw no vertices\n");
	}
#endif

#if FE_ST_TREE_DRAW_NORMALS
	if(m_vertices)
	{
		const Color red(1.0f,0.0f,0.0f,1.0f);
		const Real normalScale=1.0;
		SpatialVector pLines[2];
		for(U32 m=0;m<m_vertices;m++)
		{
			pLines[0]=m_pVertexArray[m];
			pLines[1]=pLines[0]+m_pNormalArray[m]*normalScale;

			const SpatialVector* pNormals=NULL;
			const BWORD multicolor=FALSE;
			if(a_transformed)
			{
				a_spDrawI->drawTransformedLines(*a_pTransform,NULL,pLines,
						pNormals,2,DrawI::e_discrete,multicolor,&red);
			}
			else
			{
				a_spDrawI->drawLines(pLines,pNormals,2,DrawI::e_discrete,
						multicolor,&red);
			}
		}
	}
#endif

#if FE_ST_TREE_DRAW_DEBUG
	a_spDrawI->popDrawMode();
#endif

#if FE_ST_TREE_DRAW_DEBUG
	if(m_spSpatialTree.isValid())
	{
		m_spSpatialTree->draw(a_spDrawI,NULL);
	}
#endif
}

void SurfaceTriangles::Impact::copy(sp<SpatialTreeI::Hit>& a_rspHit)
{
	SurfaceSearchable::Impact::copy(a_rspHit);
	setFace(a_rspHit->face());

	const I32* pPointIndex=a_rspHit->pointIndex();
	setPointIndex0(pPointIndex[0]);
	setPointIndex1(pPointIndex[1]);
	setPointIndex2(pPointIndex[2]);

	const SpatialVector* pVertex=a_rspHit->vertex();
	setVertex0(pVertex[0]);
	setVertex1(pVertex[1]);
	setVertex2(pVertex[2]);

	const SpatialVector* pNormal=a_rspHit->normal();
	setNormal0(pNormal[0]);
	setNormal1(pNormal[1]);
	setNormal2(pNormal[2]);
}

void SurfaceTriangles::Impact::draw(const SpatialTransform& a_rTransform,
	sp<DrawI> a_spDrawI,const fe::Color* a_pColor) const
{
	const Real inset=0.1;	//* TODO DrawMode param
	const Real scale=1.0-2.0*inset;
	const Real normalLength=0.03*
			(magnitude(m_vertex1-m_vertex0)+
			magnitude(m_vertex2-m_vertex1)+
			magnitude(m_vertex0-m_vertex2));

	SpatialVector line[4];
	line[0]=scale*m_vertex0+inset*m_vertex1+inset*m_vertex2;
	line[1]=inset*m_vertex0+scale*m_vertex1+inset*m_vertex2;
	line[2]=inset*m_vertex0+inset*m_vertex1+scale*m_vertex2;
	line[3]=line[0];

	a_spDrawI->drawTransformedLines(a_rTransform,NULL,
			line,NULL,4,DrawI::e_strip,false,a_pColor);

#if TRUE
	Color black(0.0,0.0,0.0);
	black[3]=a_pColor[0][3]*0.3;
	line[0]=0.333*(m_vertex0+m_vertex1+m_vertex2);
	line[1]=m_intersection;
	a_spDrawI->drawTransformedLines(a_rTransform,NULL,
			line,NULL,2,DrawI::e_strip,false,&black);

	Color reverse=Color(1.0,1.0,1.0)-a_pColor[0];
	reverse[3]=a_pColor[0][3];
	line[0]=line[1]+normalLength*m_normal;
	a_spDrawI->drawTransformedLines(a_rTransform,NULL,
			line,NULL,2,DrawI::e_strip,false,&reverse);
#endif
}

void SurfaceTriangles::resolveImpact(sp<ImpactI> a_spImpactI) const
{
	sp<Impact> spImpact(a_spImpactI);

	// TEMP if no nearest
	if(!spImpact.isValid() || spImpact->distance()<Real(0))
	{
		return;
	}

	SpatialVector intersection;
	SpatialVector normal;
	RayTriangleIntersect<Real>::resolveContact(
			spImpact->vertex0(),
			spImpact->vertex1(),
			spImpact->vertex2(),
			spImpact->normal0(),
			spImpact->normal1(),
			spImpact->normal2(),
			spImpact->origin(),
			spImpact->direction(),
			spImpact->barycenter(),
			spImpact->distance(),
			intersection,normal);
	spImpact->setIntersectionLocal(intersection);
	spImpact->setNormalLocal(normal);
}

} // namespace
